#include "ImagePyramidRenderer.hpp"
#include "FAST/Exception.hpp"
#include "FAST/DeviceManager.hpp"
#include "FAST/Utility.hpp"
#include "FAST/SceneGraph.hpp"
#include <FAST/Data/ImagePyramid.hpp>
#include <QGLContext>
#include <FAST/Visualization/Window.hpp>
#include <FAST/Visualization/View.hpp>
#if defined(__APPLE__) || defined(__MACOSX)
#include <OpenCL/cl_gl.h>
#include <OpenGL/gl.h>
#include <OpenGL/OpenGL.h>
#else
#if _WIN32
#include <GL/gl.h>
#include <CL/cl_gl.h>
#else
#include <CL/cl_gl.h>

#endif
#endif


namespace fast {

void ImagePyramidRenderer::clearPyramid() {
    // Clear buffer. Useful when processing a new image
    // Delete all GL data
    for(auto item : mVAO) {
        glDeleteVertexArrays(1, &item.second);
    }
    mVAO.clear();
    for(auto item : mVBO) {
        glDeleteBuffers(1, &item.second);
    }
    mVBO.clear();
    for(auto item : mEBO) {
        glDeleteBuffers(1, &item.second);
    }
    mEBO.clear();
    for(auto texture : mTexturesToRender) {
        glDeleteTextures(1, &texture.second);
    }
    mTexturesToRender.clear();
    mDataToRender.clear();
}

ImagePyramidRenderer::~ImagePyramidRenderer() {
    m_stop = true;
    m_queueEmptyCondition.notify_one();
    m_bufferThread->join();
    reportInfo() << "Buffer thread in ImagePyramidRenderer stopped" << reportEnd();
    clearPyramid();
}

ImagePyramidRenderer::ImagePyramidRenderer() : Renderer() {
    createInputPort<ImagePyramid>(0, false);
    mIsModified = true;
    m_stop = false;
    m_currentLevel = -1;
    createFloatAttribute("window", "Intensity window", "Intensity window", -1);
    createFloatAttribute("level", "Intensity level", "Intensity level", -1);
    createShaderProgram({
                                Config::getKernelSourcePath() + "/Visualization/ImagePyramidRenderer/ImagePyramidRenderer.vert",
                                Config::getKernelSourcePath() + "/Visualization/ImagePyramidRenderer/ImagePyramidRenderer.frag",
                        });

}


void ImagePyramidRenderer::loadAttributes() {
}

void ImagePyramidRenderer::draw(Matrix4f perspectiveMatrix, Matrix4f viewingMatrix, float zNear, float zFar, bool mode2D) {
    if(mDataToRender.empty())
        return;

    if(!m_bufferThread) {
        // Create thread to load patches
#ifdef WIN32
        // Create a GL context for the thread which is sharing with the context of the view
        auto context = new QGLContext(View::getGLFormat(), m_view);
        context->create(m_view->context());

        if(!context->isValid())
            throw Exception("The custom Qt GL context is invalid!");

        if(!context->isSharing())
            throw Exception("The custom Qt GL context is not sharing!");

        context->makeCurrent();
        auto nativeContextHandle = wglGetCurrentContext();
        context->doneCurrent();
        m_view->context()->makeCurrent();
        auto dc = wglGetCurrentDC();
        
        m_bufferThread = std::make_unique<std::thread>([this, dc, nativeContextHandle]() {
            wglMakeCurrent(dc, nativeContextHandle);
#else
        m_bufferThread = std::make_unique<std::thread>([this]() {
            // Create a GL context for the thread which is sharing with the context of the view
            auto context = new QGLContext(View::getGLFormat(), m_view);
            context->create(m_view->context());
            if(!context->isValid())
                throw Exception("The custom Qt GL context is invalid!");

            if(!context->isSharing())
                throw Exception("The custom Qt GL context is not sharing!");
            context->makeCurrent();
#endif
            uint64_t memoryUsage = 0;
            while(true) {
                std::string tileID;
                {
                    std::unique_lock<std::mutex> lock(m_tileQueueMutex);
                    // If queue is empty, we wait here
                    while(m_tileQueue.empty() && !m_stop) {
                        m_queueEmptyCondition.wait(lock);
                    }
                    if(m_stop)
                        break;

                    // Get next item on queue
                    tileID = m_tileQueue.back();
                    m_tileQueue.pop_back();
                }

                // Check if tile has been processed before
                if(mTexturesToRender.count(tileID) > 0)
                    continue;

                //std::cout << "Loading tile " << tileID << " queue size: " << m_tileQueue.size() << std::endl;

                // Create texture
                auto parts = split(tileID, "_");
                if(parts.size() != 3)
                    throw Exception("incorrect tile format");

                int level = std::stoi(parts[0]);
                int tile_x = std::stoi(parts[1]);
                int tile_y = std::stoi(parts[2]);
                //std::cout << "Creating texture for tile " << tile_x << " " << tile_y << " at level " << level << std::endl;
                ImagePyramidPatch tile;
                {
                    auto access = m_input->getAccess(ACCESS_READ);
                    tile = access->getPatch(level, tile_x, tile_y);
                }
                // Copy data from CPU to GL texture
                GLuint textureID;
                glGenTextures(1, &textureID);
                glBindTexture(GL_TEXTURE_2D, textureID);
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

                // TODO Why is this needed:
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);

                // WSI data from openslide is stored as ARGB, need to handle this here: BGRA and reverse
                if(m_input->isBGRA()) {
                    glTexImage2D(GL_TEXTURE_2D, 0, GL_COMPRESSED_RGBA, tile.width, tile.height, 0, GL_BGRA,
                                 GL_UNSIGNED_BYTE,
                                 tile.data.get());
                } else {
                    if(m_input->getNrOfChannels() == 3) {
                        glTexImage2D(GL_TEXTURE_2D, 0, GL_COMPRESSED_RGB, tile.width, tile.height, 0, GL_RGB,
                                     GL_UNSIGNED_BYTE,
                                     tile.data.get());
                    } else if(m_input->getNrOfChannels() == 4) {
                        glTexImage2D(GL_TEXTURE_2D, 0, GL_COMPRESSED_RGBA, tile.width, tile.height, 0, GL_RGBA,
                                     GL_UNSIGNED_BYTE,
                                     tile.data.get());
                    }
                }
                GLint compressedImageSize = 0;
                glGetTexLevelParameteriv(GL_TEXTURE_2D, 0, GL_TEXTURE_COMPRESSED_IMAGE_SIZE, &compressedImageSize);
                glBindTexture(GL_TEXTURE_2D, 0);
                glFinish();

                {
                    std::lock_guard<std::mutex> lock(m_tileQueueMutex);
					mTexturesToRender[tileID] = textureID;
                }
                memoryUsage += compressedImageSize;
                //std::cout << "Texture cache in ImagePyramidRenderer using " << (float)memoryUsage / (1024 * 1024) << " MB" << std::endl;
            }
        });
    }
    std::lock_guard<std::mutex> lock(mMutex);

    Vector4f bottom_left = (perspectiveMatrix*viewingMatrix).inverse()*Vector4f(-1,-1,0,1);
    Vector4f top_right = (perspectiveMatrix*viewingMatrix).inverse()*Vector4f(1,1,0,1);
    float width = top_right.x() - bottom_left.x();
    float height = std::fabs(top_right.y() - bottom_left.y());
    //std::cout << "Viewing coordinates:" << bottom_left.transpose() << " " <<  top_right.transpose() << std::endl;
    //std::cout << "Current Size:" << width << " " <<  height << std::endl;
    int offset_x = bottom_left.x();
    int offset_y = top_right.y();
    //std::cout << "Offset x:" << offset_x << std::endl;
    //std::cout << "Offset y:" << offset_y << std::endl;

    m_input = std::static_pointer_cast<ImagePyramid>(mDataToRender[0]);
    // Determine which level to use
    // If nr of pixels in viewport is larger than the current width and height of view, than increase the magnification
    int fullWidth = m_input->getFullWidth();
    int fullHeight = m_input->getFullHeight();
    int levelToUse = 0;
    int level = m_input->getNrOfLevels();
    do {
        level = level - 1;
        int levelWidth = m_input->getLevelWidth(level);
        int levelHeight = m_input->getLevelHeight(level);

        // Percentage of full WSI shown currently
        float percentageShownX = (float)width / fullWidth;
        float percentageShownY = (float)height / fullHeight;
        // With current level, do we have have enough pixels to fill the view?
        if(percentageShownX * levelWidth > m_view->width() && percentageShownY * levelHeight > m_view->height()) {
            // If yes, stop here
            levelToUse = level;
            break;
        } else {
            // If not, increase the magnification 
            continue;
        }
    } while(level > 0);
    if(m_currentLevel != levelToUse && m_currentLevel != -1) {
        // Level change, clear queue
        std::lock_guard<std::mutex> lock(m_tileQueueMutex);
        m_tileQueue.clear();
        //std::cout << "Queue cleared!" << std::endl;
    }
    m_currentLevel = levelToUse;

    const Vector3f spacing = m_input->getSpacing();
    activateShader();

    // This is the actual rendering
    AffineTransformation::pointer transform;
    // If rendering is in 2D mode we skip any transformations
    transform = AffineTransformation::New();

    //transform->getTransform().scale(it.second->getSpacing());

    uint transformLoc = glGetUniformLocation(getShaderProgram(), "transform");
    glUniformMatrix4fv(transformLoc, 1, GL_FALSE, transform->getTransform().data());
    transformLoc = glGetUniformLocation(getShaderProgram(), "perspectiveTransform");
    glUniformMatrix4fv(transformLoc, 1, GL_FALSE, perspectiveMatrix.data());
    transformLoc = glGetUniformLocation(getShaderProgram(), "viewTransform");
    glUniformMatrix4fv(transformLoc, 1, GL_FALSE, viewingMatrix.data());

    for(int level = m_input->getNrOfLevels()-1; level >= levelToUse; level--) {
        const int levelWidth = m_input->getLevelWidth(level);
        const int levelHeight = m_input->getLevelHeight(level);
        const int mTilesX = m_input->getLevelTilesX(level);
        const int mTilesY = m_input->getLevelTilesY(level);
        const int tileWidth = m_input->getLevelTileWidth(level);
        const int tileHeight = m_input->getLevelTileHeight(level);
        const float mCurrentTileScale = (float)fullWidth/levelWidth;

        for(int tile_x = 0; tile_x < mTilesX; ++tile_x) {
            for(int tile_y = 0; tile_y < mTilesY; ++tile_y) {
                const std::string tileString =
                        std::to_string(level) + "_" + std::to_string(tile_x) + "_" + std::to_string(tile_y);

                float tile_offset_x = tile_x * tileWidth;
                float tile_offset_y = tile_y * tileHeight;

                float tile_width = tileWidth;
                if(tile_x == mTilesX - 1)
                    tile_width = levelWidth - tile_offset_x;
                float tile_height = tileHeight;
                if(tile_y == mTilesY - 1)
                    tile_height = levelHeight - tile_offset_y;

                //tile_width *= spacing.x();
                //tile_height *= spacing.y();
                //tile_offset_x *= spacing.x();
                //tile_offset_y *= spacing.y();

                // Only process visible patches
                // Fully contained and partly
                if(!(
                        (offset_x <= tile_offset_x * mCurrentTileScale &&
                         offset_x + width > tile_offset_x * mCurrentTileScale + tile_width * mCurrentTileScale)
                        ||
                        (offset_x > tile_offset_x * mCurrentTileScale &&
                         offset_x < (tile_offset_x + tile_width) * mCurrentTileScale)
                        ||
                        (offset_x + width > tile_offset_x * mCurrentTileScale &&
                         offset_x + width < (tile_offset_x + tile_width) * mCurrentTileScale)
                ))
                    continue;
                if(!(
                        (offset_y <= tile_offset_y * mCurrentTileScale &&
                         offset_y + height > tile_offset_y * mCurrentTileScale + tile_height * mCurrentTileScale)
                        ||
                        (offset_y > tile_offset_y * mCurrentTileScale &&
                         offset_y < (tile_offset_y + tile_height) * mCurrentTileScale)
                        ||
                        (offset_y + height > tile_offset_y * mCurrentTileScale &&
                         offset_y + height < (tile_offset_y + tile_height) * mCurrentTileScale)
                ))
                    continue;


                // Is patch in cache?
                bool textureReady = false;
                uint textureID;
                {
					std::lock_guard<std::mutex> lock(m_tileQueueMutex);
                    textureReady = mTexturesToRender.count(tileString) > 0;
                }
                if(!textureReady) {
                    // Add to queue if not in cache
                    {
                        std::lock_guard<std::mutex> lock(m_tileQueueMutex);
                        // Remove any duplicates first
                        m_tileQueue.remove(tileString); // O(n) time complexity..
						m_tileQueue.push_back(tileString);
                        //std::cout << "Added tile " << tileString << " to queue" << std::endl;
                    }
                    m_queueEmptyCondition.notify_one();
                    continue;
                } else {
                    textureID = mTexturesToRender[tileString];
                }

                // Delete old VAO
                if(mVAO.count(tileString) == 0) {
                    // Create VAO
                    uint VAO_ID;
                    glGenVertexArrays(1, &VAO_ID);
                    mVAO[tileString] = VAO_ID;
                    glBindVertexArray(VAO_ID);

                    // Create VBO
                    // Get width and height in mm
                    //std::cout << "Creating vertices for " << tile_x << " " << tile_y << std::endl;
                    //std::cout << "Tile position: " << tile_offset_x*mCurrentTileScale << " " << tile_offset_x*mCurrentTileScale + tile_width*mCurrentTileScale << std::endl;
                    //std::cout << "Tile position: " << tile_offset_y*mCurrentTileScale << " " << tile_offset_y*mCurrentTileScale + tile_height*mCurrentTileScale << std::endl;
                    float vertices[] = {
                            // vertex: x, y, z; tex coordinates: x, y
                            tile_offset_x * mCurrentTileScale, (tile_offset_y + tile_height) * mCurrentTileScale,
                            -level,
                            0.0f, 1.0f,
                            (tile_offset_x + tile_width) * mCurrentTileScale,
                            (tile_offset_y + tile_height) * mCurrentTileScale, -level, 1.0f, 1.0f,
                            (tile_offset_x + tile_width) * mCurrentTileScale, tile_offset_y * mCurrentTileScale, -level,
                            1.0f,
                            0.0f,
                            tile_offset_x * mCurrentTileScale, tile_offset_y * mCurrentTileScale, -level, 0.0f, 0.0f,
                    };
                    uint VBO;
                    glGenBuffers(1, &VBO);
                    mVBO[tileString] = VBO;
                    glBindBuffer(GL_ARRAY_BUFFER, VBO);
                    glBufferData(GL_ARRAY_BUFFER, sizeof(vertices), vertices, GL_STATIC_DRAW);
                    glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 5 * sizeof(float), (void *) 0);
                    glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, 5 * sizeof(float), (void *) (3 * sizeof(float)));
                    glEnableVertexAttribArray(0);
                    glEnableVertexAttribArray(1);

                    // Create EBO
                    uint EBO;
                    glGenBuffers(1, &EBO);
                    mEBO[tileString] = EBO;
                    uint indices[] = {  // note that we start from 0!
                            0, 1, 3,   // first triangle
                            1, 2, 3    // second triangle
                    };
                    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, EBO);
                    glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(indices), indices, GL_STATIC_DRAW);
                    glBindVertexArray(0);
                }

                //std::cout << "drawing " << tileString << std::endl;
                glBindTexture(GL_TEXTURE_2D, textureID);
                glBindVertexArray(mVAO[tileString]);
                glDrawElements(GL_TRIANGLES, 6, GL_UNSIGNED_INT, 0);
                glBindTexture(GL_TEXTURE_2D, 0);
                glBindVertexArray(0);
                glFinish();
            }
        }
    }
    deactivateShader();
}

void ImagePyramidRenderer::drawTextures(Matrix4f &perspectiveMatrix, Matrix4f &viewingMatrix, bool mode2D) {

}


} // end namespace fast
